# Chapter 15: Output Hotplugging and Seat

In our earlier discussion within [Chapter 4: Compositor Initialization](md_md_tutorial_04.html), we explored the arrangement of outputs during the compositor initialization process. This chapter delves into the scenarios involving the connection and disconnection of outputs, and as a final addition, we will delve into the handling of TTY switch events.

Whenever a new output is connected, the Louvre::LSeat::outputPlugged() event is triggered. Within this method, you have the freedom to decide whether to initialize the output using Louvre::LCompositor::addOutput() or not. Similarly, when an output is no longer available, the Louvre::LSeat::outputUnplugged() event is triggered. If the given output was initialized while unplugged, it is automatically uninitialized after this event, even if you don't explicitly call Louvre::LCompositor::removeOutput().

Let's override these methods to see how to handle these situations:

#### src/ESeat.h

```cpp
// ...

class ESeat : public LSeat
{
public:
    // ...

    void outputPlugged(LOutput *output) override;
    void outputUnplugged(LOutput *output) override;
};

// ...
```

#### src/ESeat.cpp

```cpp
// ..

#include "EOutput.h"
#include "Global.h"

void ESeat::outputPlugged(LOutput *output)
{
    output->setScale(output->dpi() >= 200 ? 2 : 1);

    LOutput *lastOutput = compositor()->outputs().back();

    if (lastOutput)
        output->setPos(LPoint(lastOutput->pos().x() + lastOutput->size().x(), 0));
    else
        output->setPos(LPoint(0, 0));

    compositor()->addOutput(output);
    compositor()->repaintAllOutputs();
}

void ESeat::outputUnplugged(LOutput *output)
{
    compositor()->removeOutput(output);

    Int32 totalWidth = 0;

    // Organize still initialized outputs horizontally and sequentially.
    for (EOutput *o : G::outputs())
    {
        o->setPos(LPoint(totalWidth, 0));
        totalWidth += o->size().w();
        o->updateSurfaces();
    }

    compositor()->repaintAllOutputs();
}
```

When a new output is plugged, our approach is to adjust its scale factor, position it at the rightmost side of the already initialized outputs, and initialize it using Louvre::LCompositor::addOutput().

And when an output is unplugged, we first manually uninitialize it just for removing it from the Louvre::LCompositor::outputs() list. Afterward, we re-arrange the still-initialized outputs and call the `updateSurfaces()` method we introduced earlier. This serves the dual purpose of updating any maximized or fullscreen toplevel surfaces and ensuring that surfaces remain accessible to the user.

For a firsthand experience of these changes, recompile and run the compositor, and experiment with plugging and unplugging outputs on the fly to observe how the compositor reacts.

### TTY Switching

Throughout this tutorial, we've frequently referred to the Louvre::LSeat class, and you may be curious about its purpose. In the realm of Linux/Wayland, the concept of a "seat" represents a collection of input and output devices. Think of it as a physical "seat" where a user is situated, providing access to their keyboard, mouse, screen, and more. This concept becomes particularly important in scenarios like educational institutions where multiple students can share the same machine, while each having their dedicated monitor, keyboard, and so on. This setup is commonly known as a multi-seat configuration.

However, in the most typical scenario, a single machine is equipped with a single set of input and output devices. In such cases, Linux still allows you to run multiple user sessions concurrently but in different virtual terminals (TTYs), which you can switch between using shortcuts like `Ctrl + Shift + [F1, F2, ..., FN]` or through the `chvt` command. This capability enables you to have multiple desktop environments running simultaneously or, in the context of Louvre, run multiple compositors on different TTYs.

Each of these desktop environments, compositors, etc., needs to interact with input devices and render content on the screen. However, if there is only one physical screen, this simultaneous interaction would be chaotic. Linux device drivers are designed to allow only one process to manage them at a given time. For example, in the DRM/KMS API, the concept of the DRM master exists, and only one process can be the DRM master, thereby controlling the screen content.

Since there are multiple TTYs, there needs to be a way to negotiate these permissions during TTY switches. This is where seat services and libraries like [libseat](https://github.com/kennylevinsen/seatd) come into play. Programs that integrate with seat services enable other processes to take ownership of the seat, and Louvre is no exception.

By default, Louvre allows users to switch between TTYs using the methods mentioned earlier or by invoking the Louvre::LSeat::setTTY() method. When the user switches to another TTY, the compositor can no longer listen to input events or draw content on screens. You have the option to disable this feature by setting the **LOUVRE_ENABLE_LIBSEAT** environment variable to 0. When the compositor is run in this manner, it won't allow any other process to take ownership of the seat, and TTY switching won't be possible.

With this feature enabled and while the user is in another TTY, other processes have the potential to modify resources used by the compositor, such as the hardware cursor plane and CRTCs, among others. To properly handle those cases and keep track of TTY switches, you can override the Louvre::LSeat::enabledChanged() virtual method.

#### src/ESeat.h

```cpp
// ...

class ESeat : public LSeat
{
public:
    // ...

    // TTY switching
    void enabledChanged() override;
};

// ...
```

#### src/ESeat.cpp

```cpp
// ..

#include <LSceneView.h>
#include <LScene.h>
#include <LCursor.h>

// ...

void ESeat::enabledChanged()
{
    if (!enabled())
        return;

    /* Synchronize LScene's output inter-frame damage tracking by performing
     * full damage on all outputs. */
    for (EOutput *output : G::outputs())
    {
        G::scene()->mainView()->damageAll(output);
        output->repaint();
    }

    /* Ensure the graphic backend restores the hardware cursor's state,
     * which may have changed in another session. */
    cursor()->setVisible(false);
    cursor()->setVisible(true);
    cursor()->move(1,1);
}
```

Here, every time the seat is re-enabled (when a user returns from another session), we apply damage to all the outputs within our scene. We do this as the inter-frame damage of each output could have fallen out of sync, potentially leading to odd visual glitches.

We also toggle the cursor's visibility and move it slightly to ensure that the hardware cursor's state is restored to where it was left.

### Congratulations!

We've covered the most crucial aspects of building a compositor with Louvre, and you may already be prepared to let your creativity soar. I'm excited to see what you can create.

If you encounter any bugs, have suggestions, questions, want to share your creations or contribute to the development of Louvre, please don't hesitate to get in touch. You can find our contact information on the [Contact](md_md__contact.html) page.

<a href="md_md_tutorial_14.html">◀ Chapter 14: Clipboard and DND</a>